--******************************************************************************
--
--  Xilinx, Inc. 2004                 www.xilinx.com
--
--  The Swiss Army Knife LUT
--
--*******************************************************************************
--
--  File name :       LUT.VHD
--
--  Description :     This is a very generic LUT instantiation component, the Swiss Army Knife LUT 
--
--                    Using only one component you can instantiate LUT1, LUT2, LUT3 or LUT4 primitives, 
--                    define their INIT property as a logic equation string instead of a BIT_VECTOR, 
--                    build RPMs with RLOC and BEL properties, place directed routing constraints on 
--                    the LUT output and do timing accurate simulation of asynchronous logic built 
--                    with it in functional simulation mode 
--
--                    It has been tested with Synplicity SynplifyPro, Mentor Graphics Precision and ModelSim
--
--  Date :            03/05/2004
--
--  Revision :        1.2
--
--  Author :          Catalin Baetoniu
--
--  Contact :         catalinb@xilinx.com - This design is not supported by the hotline
--
--  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
--              provided to you "as is". Xilinx and its licensors make and you 
--              receive no warranties or conditions, express, implied, 
--              statutory or otherwise, and Xilinx specifically disclaims any 
--              implied warranties of merchantability, non-infringement, or 
--              fitness for a particular purpose. Xilinx does not warrant that 
--              the functions contained in these designs will meet your 
--              requirements, or that the operation of these designs will be 
--              uninterrupted or error free, or that defects in the Designs 
--              will be corrected. Furthermore, Xilinx does not warrant or 
--              make any representations regarding use or the results of the 
--              use of the designs in terms of correctness, accuracy, 
--              reliability, or otherwise. 
--
--              LIMITATION OF LIABILITY. In no event will Xilinx or its 
--              licensors be liable for any loss of data, lost profits, cost 
--              or procurement of substitute goods or services, or for any 
--              special, incidental, consequential, or indirect damages 
--              arising from the use or operation of the designs or 
--              accompanying documentation, however caused and on any theory 
--              of liability. This limitation will apply even if Xilinx 
--              has been advised of the possibility of such damage. This 
--              limitation shall apply not-withstanding the failure of the 
--              essential purpose of any limited remedies herein. 
--
--  Copyright  2004 Xilinx, Inc.
--  All rights reserved 
-- 
--*****************************************************************************

library IEEE;
use IEEE.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

entity LUT is
  generic(EQU:STRING;          -- the logic equation the LUT implements - it must be specified
          FG:STRING:="";       -- the BEL attribute, must be "F" or "G" - it is optional
          DIRT:STRING:="";     -- the DIRT string for the LUT output    - it is optional
          LP:STRING:="false";  -- the LOCK_PINS attribute               - it is optional, default is false
          DELAY:TIME:=0 ps);   -- the LUT propagation delay (Tilo)      - it is optional
  port(O:out STD_LOGIC;
       I0:in STD_LOGIC:='0';   -- to instantiate LUT1, LUT2 or LUT3
       I1:in STD_LOGIC:='0';   -- simply let the coresponding inputs
       I2:in STD_LOGIC:='0';   -- unconnected
       I3:in STD_LOGIC:='0');
end LUT;

architecture SwissArmyKnife of LUT is
  function MAGIC(EQU:STRING) return BIT_VECTOR is
    variable A:BIT_VECTOR(15 downto 0);
    variable P:INTEGER;
-- the parser implements the following recursive grammar, where B is a logic equation and A an operand:
-- B::=A|not A|~A|A {and A}|A{*A}|A {or A}|A{+A}|A {xor A}|A{@A}|A xnor A
-- A::=I0|I1|I2|I3|(B)
    procedure PARSE(FLAG:in BOOLEAN;RESULT:out BIT_VECTOR) is
      variable A,B:BIT_VECTOR(15 downto 0);
      variable CONT:BOOLEAN;
    begin
      if P<=EQU'high then
        while P<EQU'high and EQU(P)=' ' loop
          P:=P+1;
        end loop;
        if EQU(P)=' ' then
          P:=P+1;
        end if;
      end if;
      if FLAG then
        if P>EQU'high then
          RESULT:=X"0000";
        else
          if (EQU(P)='n' or EQU(P)='N') and
             (EQU(P+1)='o' or EQU(P+1)='O') and
             (EQU(P+2)='t' or EQU(P+2)='T') then
            P:=P+3;
            PARSE(FALSE,A);
            RESULT:=not A;
          elsif EQU(P)='~' then
            P:=P+1;
            PARSE(FALSE,A);
            RESULT:=not A;
          else
            PARSE(FALSE,A);
            if P<=EQU'high then
              while P<EQU'high and EQU(P)=' ' loop
                P:=P+1;
              end loop;
              if EQU(P)=' ' then
                P:=P+1;
              end if;
            end if;
            if (EQU(P)='x' or EQU(P)='X') and
               (EQU(P+1)='n' or EQU(P+1)='N') and
               (EQU(P+2)='o' or EQU(P+2)='O') and
               (EQU(P+3)='r' or EQU(P+3)='R') then
              P:=P+4;
              PARSE(FALSE,B);
              A:=A xor not B;
            elsif (EQU(P)='x' or EQU(P)='X') and
                  (EQU(P+1)='o' or EQU(P+1)='O') and
                  (EQU(P+2)='r' or EQU(P+2)='R') then
              CONT:=TRUE;
              while CONT loop
                P:=P+3;
                PARSE(FALSE,B);
                A:=A xor B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P+2<EQU'high then
                  CONT:=(EQU(P)='x' or EQU(P)='X') and
                        (EQU(P+1)='o' or EQU(P+1)='O') and
                        (EQU(P+2)='r' or EQU(P+2)='R');
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            elsif EQU(P)='@' then
              CONT:=TRUE;
              while CONT loop
                P:=P+1;
                PARSE(FALSE,B);
                A:=A xor B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P<EQU'high then
                  CONT:=EQU(P)='@';
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            elsif (EQU(P)='o' or EQU(P)='O') and
                  (EQU(P+1)='r' or EQU(P+1)='R') then
              CONT:=TRUE;
              while CONT loop
                P:=P+2;
                PARSE(FALSE,B);
                A:=A or B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P+1<EQU'high then
                  CONT:=(EQU(P)='o' or EQU(P)='O') and
                        (EQU(P+1)='r' or EQU(P+1)='R');
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            elsif EQU(P)='+' then
              CONT:=TRUE;
              while CONT loop
                P:=P+1;
                PARSE(FALSE,B);
                A:=A or B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P<EQU'high then
                  CONT:=EQU(P)='+';
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            elsif (EQU(P)='a' or EQU(P)='A') and
                  (EQU(P+1)='n' or EQU(P+1)='N') and
                  (EQU(P+2)='d' or EQU(P+2)='D') then
              CONT:=TRUE;
              while CONT loop
                P:=P+3;
                PARSE(FALSE,B);
                A:=A and B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P+2<EQU'high then
                  CONT:=(EQU(P)='a' or EQU(P)='A') and
                        (EQU(P+1)='n' or EQU(P+1)='N') and
                        (EQU(P+2)='d' or EQU(P+2)='D');
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            elsif EQU(P)='*' then
              CONT:=TRUE;
              while CONT loop
                P:=P+1;
                PARSE(FALSE,B);
                A:=A and B;
                if P<=EQU'high then
                  while P<EQU'high and EQU(P)=' ' loop
                    P:=P+1;
                  end loop;
                  if EQU(P)=' ' then
                    P:=P+1;
                  end if;
                end if;
                if P<EQU'high then
                  CONT:=EQU(P)='*';
                else
                  CONT:=FALSE;
                end if;
              end loop;  
            end if;
            RESULT:=A;
          end if;
        end if;        
      else
        case EQU(P) is
          when 'i'|'I'=>P:=P+1;
                        case EQU(P) is
                          when '0'=>RESULT:=X"AAAA";
                          when '1'=>RESULT:=X"CCCC";
                          when '2'=>RESULT:=X"F0F0";
                          when '3'=>RESULT:=X"FF00";
                          when others=>RESULT:=X"0000";
                                       assert FALSE report "" severity error;
                        end case;
                        P:=P+1;
          when '('=>P:=P+1;
                    PARSE(TRUE,A);
                    RESULT:=A;
                    while P<EQU'high and EQU(P)=' ' loop
                      P:=P+1;
                    end loop;
                    if EQU(P)=' ' then
                      P:=P+1;
                    end if;
                    assert EQU(P)=')' report "Syntax Error!" severity error;
                    P:=P+1;
          when others=>RESULT:=X"0000";
                       assert FALSE report "Syntax Error!" severity error;
        end case;
      end if;
    end;
  begin
    P:=EQU'low;
    PARSE(TRUE,A);
    assert EQU(P to EQU'high)="...." report "Syntax Error!" severity error;
    return A;
  end;
  
  constant INIT:BIT_VECTOR:=MAGIC(EQU&"...."); -- this trick solves a lot of issues
  attribute rloc:STRING;
  attribute bel:STRING;
  attribute syn_keep:BOOLEAN;
  attribute route:STRING;
  attribute lock_pins:STRING;

  signal OO:STD_LOGIC;
  attribute syn_keep of OO:signal is TRUE;
  attribute route of OO:signal is DIRT;
begin
  g4:if INIT(15 downto 8)/=INIT(7 downto 0) generate
       attribute bel of l4:label is FG;
       attribute rloc of l4:label is "X0Y0";
       attribute lock_pins of l4:label is LP;
     begin
       l4:LUT4 generic map(INIT=>INIT)
               port map(I0=>I0,I1=>I1,I2=>I2,I3=>I3,O=>OO);
     end generate;
     
  g3:if INIT(15 downto 8)=INIT(7 downto 0) and
        INIT(7 downto 4)/=INIT(3 downto 0) generate
       attribute rloc of l3:label is "X0Y0";
       attribute bel of l3:label is FG;
       attribute lock_pins of l3:label is LP;
     begin
       l3:LUT3 generic map(INIT=>INIT(7 downto 0))
               port map(I0=>I0,I1=>I1,I2=>I2,O=>OO);
     end generate;
     
  g2:if INIT(15 downto 8)=INIT(7 downto 0) and
        INIT(7 downto 4)=INIT(3 downto 0) and
        INIT(3 downto 2)/=INIT(1 downto 0) generate
       attribute rloc of l2:label is "X0Y0";
       attribute bel of l2:label is FG;
       attribute lock_pins of l2:label is LP;
     begin
       l2:LUT2 generic map(INIT=>INIT(3 downto 0))
               port map(I0=>I0,I1=>I1,O=>OO);
     end generate;
     
  g1:if INIT(15 downto 8)=INIT(7 downto 0) and
        INIT(7 downto 4)=INIT(3 downto 0) and
        INIT(3 downto 2)=INIT(1 downto 0) generate
       attribute rloc of l1:label is "X0Y0";
       attribute bel of l1:label is FG;
       attribute lock_pins of l1:label is LP;
     begin
       l1:LUT1 generic map(INIT=>INIT(1 downto 0))
               port map(I0=>I0,O=>OO);
     end generate;

  O<=transport OO after DELAY;
end SwissArmyKnife;
